import logging
from pathlib import Path

from rich.logging import RichHandler


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    handler = RichHandler()
    logger.addHandler(handler)
    logger.propagate = False
    return logger


def get_cache_dir() -> Path:
    cache_dir = Path.home() / ".cache" / "tgpt"
    cache_dir.mkdir(exist_ok=True, parents=True)
    return cache_dir



from functools import cache

from currency_converter import CurrencyConverter


@cache
def get_rate(from_currency: str, to_currency: str) -> float:
    converter = CurrencyConverter()
    rate = converter.convert(1, from_currency, to_currency)
    assert isinstance(rate, float)
    return rate


def usd_to_gbp(usd: float) -> float:
    rate = get_rate("USD", "GBP")
    gbp = usd * rate
    return gbp



import datetime
import os
import tempfile
import time
from pathlib import Path

import openai
from pydub import AudioSegment
from telegram import Update
from telegram import Voice
from telegram.ext import ApplicationBuilder
from telegram.ext import CommandHandler
from telegram.ext import ContextTypes
from telegram.ext import MessageHandler
from telegram.ext import filters

from .apis import GPT4
from .apis import Whisper
from .utils import get_cache_dir
from .utils import get_logger

logger = get_logger(__name__)


class Bot:
    def __init__(self) -> None:
        logger.info("Initializing bot")
        self._transcriber = Whisper()
        self._chatbot_class = GPT4
        self._chatbot = self._chatbot_class()
        telegram_token = os.environ["TELEGRAM_BOT_TOKEN"]
        self._application = ApplicationBuilder().token(telegram_token).build()
        self._add_handlers()
        self._cache_dir = get_cache_dir()
        self._show_cost = False

    def _add_handlers(self) -> None:
        text_handler = MessageHandler(
            filters.TEXT & ~filters.COMMAND,
            self._text_callback,
        )
        self._application.add_handler(text_handler)

        voice_handler = MessageHandler(filters.VOICE, self._voice_callback)
        self._application.add_handler(voice_handler)

        new_session_handler = CommandHandler(
            "new",
            self._start_session,
        )
        self._application.add_handler(new_session_handler)

    async def _chat(self, text: str) -> tuple[str, float]:
        logger.info("Getting response from chatbot...")
        await self._send("[Getting response from chatbot...]")
        reply, cost = self._chatbot(text)
        logger.info(f'Reply received: "{reply}"')
        logger.info(f"Cost: £{100 * cost:.3}p")
        return reply, cost

    async def _send(self, message: str) -> None:
        assert self.update.effective_chat is not None
        await self.context.bot.send_message(
            chat_id=self.update.effective_chat.id,
            text=message,
        )

    def _set_update_context(
        self,
        update: Update,
        context: ContextTypes.DEFAULT_TYPE,
    ) -> None:
        self.update = update
        self.context = context

    async def _text_callback(
        self,
        update: Update,
        context: ContextTypes.DEFAULT_TYPE,
    ) -> None:
        self._set_update_context(update, context)
        if not await self._check_is_me():
            return
        assert update.message is not None
        text = update.message.text
        assert isinstance(text, str)
        logger.info(f'Text message received: "{text}"')
        reply, cost = await self._chat(text)
        assert update.effective_chat is not None
        if self._show_cost:
            await self._send(f"[Cost: £{100 * cost:.3}p]")
        await self._send(reply)

    async def _check_is_me(self) -> bool:
        my_client_id = int(os.environ["TELEGRAM_CLIENT_ID"])
        assert self.update.effective_chat is not None
        is_me = self.update.effective_chat.id == my_client_id
        if is_me:
            return True
        else:
            await self._send("Sorry, I'm not allowed to chat with you.")
            return False

    async def _start_session(
        self,
        update: Update,
        context: ContextTypes.DEFAULT_TYPE,
    ) -> None:
        self._set_update_context(update, context)
        if not await self._check_is_me():
            return
        message = f"Initializing chatbot session ({self._chatbot_class.__name__})"
        logger.info(message)
        _ = self._send(message)
        self._chatbot = self._chatbot_class()

    async def _get_mp3_from_voice(self, voice: Voice) -> Path:
        logger.info(f"Voice message received ({voice.duration} seconds)")
        voice_file = await voice.get_file()
        time_string = datetime.datetime.now().isoformat()
        filename = f"{time_string}_voice.mp3"
        mp3_path = self._cache_dir / filename
        with tempfile.NamedTemporaryFile(suffix=".oga") as f:
            oga_path = Path(f.name)
            await voice_file.download_to_drive(oga_path)
            oga_file: AudioSegment = AudioSegment.from_file(oga_path, format="ogg")
            oga_file.export(mp3_path, format="mp3")
        return mp3_path

    async def _voice_callback(
        self,
        update: Update,
        context: ContextTypes.DEFAULT_TYPE,
        max_num_attempts: int = 3,
        waiting_time: float = 1,
    ) -> None:
        self._set_update_context(update, context)
        if not await self._check_is_me():
            return

        assert update.message is not None
        assert isinstance(update.message.voice, Voice)
        voice = update.message.voice
        logger.info(f"Voice message received ({voice.duration} seconds)")
        mp3_path = await self._get_mp3_from_voice(voice)

        logger.info("Transcribing audio...")
        assert update.effective_chat is not None
        await self._send(f"[Transcribing {voice.duration} seconds of audio...]")
        attempts = 0
        while attempts < max_num_attempts:
            try:
                transcript = self._transcriber(mp3_path)
                logger.info(f'Transcript: "{transcript}"')
                await self._send(f'[Transcript: "{transcript}"]')
                reply, cost = await self._chat(transcript)
                if self._show_cost:
                    await self._send(f"[Cost: £{100 * cost:.3}p]")
                await self._send(reply)
                return
            except openai.error.APIConnectionError:
                await self._send("[Transcription error. Trying again...]")
                attempts += 1
                time.sleep(waiting_time)
        await self._send("[Transcription API not working. Try using text...]")

    def run(self) -> None:
        logger.info("The bot is ready")
        self._application.run_polling()



from dataclasses import asdict
from dataclasses import dataclass

import openai

from .exchange import usd_to_gbp


@dataclass
class SystemMessage:
    content: str
    role: str = "system"


@dataclass
class UserMessage:
    content: str
    role: str = "user"


@dataclass
class AssistantMessage:
    content: str
    role: str = "assistant"


class GPT:
    _MODEL_NAME: str
    _PRICING_USD: float
    _DEFAULT_SYSTEM_MESSAGE = (
        "You are ChatGPT, a large language model trained by OpenAI."
        " Answer as concisely as possible."
    )

    def __init__(self) -> None:
        self.history: list[SystemMessage | UserMessage | AssistantMessage]
        self.history = [SystemMessage(self._DEFAULT_SYSTEM_MESSAGE)]

    @property
    def gbp_per_token(self) -> float:
        return usd_to_gbp(self._PRICING_USD)

    def cost_gbp(self, num_tokens: int) -> float:
        return num_tokens * self.gbp_per_token

    @property
    def _messages(self) -> list[dict[str, str]]:
        return [asdict(message) for message in self.history]

    def _call_api(self) -> openai.openai_object.OpenAIObject:
        completion = openai.ChatCompletion.create(  # type: ignore[no-untyped-call]
            model=self._MODEL_NAME,
            messages=self._messages,
        )
        assert isinstance(completion, openai.openai_object.OpenAIObject)
        return completion

    def __call__(self, message: str) -> tuple[str, float]:
        self.history.append(UserMessage(message))
        completion = self._call_api()
        reply = completion.choices[0].message.content
        self.history.append(AssistantMessage(reply))
        num_tokens = completion.usage.total_tokens
        cost_gbp = self.cost_gbp(num_tokens)
        return reply, cost_gbp


class ChatGPT(GPT):
    _MODEL_NAME = "gpt-3.5-turbo"
    _PRICING_USD = 0.002 / 1000  # $0.002 per 1000 tokens


class GPT4(GPT):
    _MODEL_NAME = "gpt-4"
    _PRICING_USD = 0.03 / 1000  # $0.03 per 1000 tokens



from pathlib import Path

import openai

from ..utils import get_logger
from .exchange import usd_to_gbp

logger = get_logger(__name__)


class Whisper:
    _PRICE_DOLLARS_PER_SECOND = 0.006 / 60
    _MODEL = "whisper-1"
    _PROMPT = (
        "The transcript is always written in English,"
        " independently of the input language."
    )

    def __init__(self) -> None:
        self.transcribe = openai.Audio.transcribe
        self.translate = openai.Audio.translate

    def get_price_pounds(self, seconds: float) -> float:
        dollars = seconds * self._PRICE_DOLLARS_PER_SECOND
        gbp = usd_to_gbp(dollars)
        pence = gbp * 100
        logger.info(f"Estimated transcription cost: {pence:.2f}p")
        return gbp

    def __call__(self, path: Path, translate: bool = True) -> str:
        process = self.translate if translate else self.transcribe
        with open(path, "rb") as f:
            transcript = process(  # type: ignore[no-untyped-call]
                self._MODEL,
                f,
                prompt=self._PROMPT,
            )
        assert isinstance(transcript.text, str)
        return transcript.text
